package com.jakubu9333.grammarManager.service;

import edu.cmu.sphinx.jsgf.JSGFGrammarException;
import edu.cmu.sphinx.jsgf.JSGFGrammarParseException;
import edu.cmu.sphinx.jsgf.JSGFRuleGrammar;
import edu.cmu.sphinx.jsgf.JSGFRuleGrammarFactory;
import edu.cmu.sphinx.jsgf.JSGFRuleGrammarManager;
import edu.cmu.sphinx.jsgf.parser.JSGFParser;
import edu.cmu.sphinx.jsgf.parser.TokenMgrError;
import edu.cmu.sphinx.jsgf.rule.JSGFRule;
import edu.cmu.sphinx.jsgf.rule.JSGFRuleAlternatives;
import edu.cmu.sphinx.jsgf.rule.JSGFRuleCount;
import edu.cmu.sphinx.jsgf.rule.JSGFRuleName;
import edu.cmu.sphinx.jsgf.rule.JSGFRuleSequence;
import edu.cmu.sphinx.jsgf.rule.JSGFRuleTag;
import edu.cmu.sphinx.jsgf.rule.JSGFRuleToken;
import org.jvoicexml.processor.grammar.Rule;
import org.jvoicexml.processor.grammar.RuleAlternatives;
import org.jvoicexml.processor.grammar.RuleComponent;
import org.jvoicexml.processor.grammar.RuleCount;
import org.jvoicexml.processor.grammar.RuleReference;
import org.jvoicexml.processor.grammar.RuleSequence;
import org.jvoicexml.processor.grammar.RuleSpecial;
import org.jvoicexml.processor.grammar.RuleTag;
import org.jvoicexml.processor.grammar.RuleToken;
import org.jvoicexml.processor.srgs.GrammarException;
import org.jvoicexml.processor.srgs.xml.SrgsRuleGrammarParser;
import org.springframework.http.HttpStatus;
import org.springframework.stereotype.Service;
import org.springframework.web.server.ResponseStatusException;
import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import java.io.IOException;
import java.io.StringReader;
import java.net.URISyntaxException;
import java.util.List;
import java.util.Map;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * @author Jakub Uhlarik
 */
@Service
public class FileService {



    public static class XmlGrammarHandler extends DefaultHandler {
        private boolean rule = false;
        private boolean grammar = false;

        @Override
        public void startElement(String uri, String lName, String qName, Attributes attr) {
            switch (qName) {
                case "grammar" -> {
                    grammar = true;
                }
                case "rule" -> {
                    if (!grammar) {
                        throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "rule outside grammar");
                    }
                    rule = true;
                }
                case "item" -> {
                    if (!rule) {
                        throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "item outside rule");
                    }
                }
                case "one-of" -> {
                    if (!rule) {
                        throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "one-of outside rule");
                    }
                }
            }
        }

        @Override
        public void endElement(String uri, String lName, String qName) {
            if ("rule".equals(qName)) {
                rule = false;
            }
        }
    }

    public  String getParsedFile(String file, Boolean xml) {
        if (xml == null) {
            xml = getFileExtension(file);

        }
        return parseFile(file, xml);
    }


    //function validates xml document and transforms jsgf document to xml
    private  String parseFile(String file, boolean xml) {

        if (xml) {
            try {
                SAXParserFactory factory = SAXParserFactory.newInstance();
                SAXParser saxParser = factory.newSAXParser();

                XmlGrammarHandler handler = new XmlGrammarHandler();
                InputSource is = new InputSource(new StringReader(file.trim()));
                saxParser.parse(is, handler);
                return file;

            } catch (ParserConfigurationException | SAXException | IOException e) {
                throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "xml parse error\n" + e.getMessage(), e);
            }

        }
        try {
            return jsgfToXml(file);
        } catch (JSGFGrammarParseException | TokenMgrError | IOException | URISyntaxException |
                 JSGFGrammarException e) {
            if (e instanceof JSGFGrammarParseException jsgfGrammarParseException) {
                throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "jsgf parse error\n" + jsgfGrammarParseException.details);
            }
            throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "jsgf parse error\n" + e.getMessage(), e);
        }


    }


    private  boolean getFileExtension(String file) {
        Pattern xml = Pattern.compile("^\\W*<\\?xml.*\\?>");
        Matcher mXml = xml.matcher(file);
        if (mXml.find()) {
            return true;
        }
        Pattern jsgf = Pattern.compile("^\\W*#JSGF.*;");
        Matcher mJsgf = jsgf.matcher(file);
        if (mJsgf.find()) {
            return false;
        }
        throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "format cant be found from file");
    }


    private static RuleComponent jsgfToXml(JSGFRule jsgfRule) {
        if (jsgfRule instanceof JSGFRuleAlternatives jsgfRuleAlternatives) {
            if (jsgfRuleAlternatives.getRules().size() == 1) {
                return jsgfToXml(jsgfRuleAlternatives.getRules().get(0));
            }
            RuleAlternatives ruleAlternatives = new RuleAlternatives();
            List<Float> weights = jsgfRuleAlternatives.getWeights();
            AtomicInteger i = new AtomicInteger();
            jsgfRuleAlternatives.getRules().forEach(p -> {
                if (weights == null) {
                    ruleAlternatives.addAlternative(jsgfToXml(p));
                } else {
                    ruleAlternatives.addAlternative(jsgfToXml(p), weights.get(i.get()));
                    i.getAndIncrement();
                }
            });
            return ruleAlternatives;
        }
        if (jsgfRule instanceof JSGFRuleSequence jsgfRuleSequence) {
            if (jsgfRuleSequence.getRules().size() == 1) {
                return jsgfToXml(jsgfRuleSequence.getRules().get(0));
            }
            RuleSequence ruleSequence = new RuleSequence();
            jsgfRuleSequence.getRules().forEach(p -> ruleSequence.addElement(jsgfToXml(p)));
            return ruleSequence;
        }
        if (jsgfRule instanceof JSGFRuleCount jsgfRuleCount) {
            int repeat = jsgfRuleCount.getCount();
            RuleComponent ruleComponent = jsgfToXml(jsgfRuleCount.getRule());
            RuleCount ruleCount;
            switch (repeat) {
                case JSGFRuleCount.ZERO_OR_MORE -> ruleCount = new RuleCount(ruleComponent, 0);
                case JSGFRuleCount.OPTIONAL -> ruleCount = new RuleCount(ruleComponent, 0, 1);
                case JSGFRuleCount.ONCE_OR_MORE -> ruleCount = new RuleCount(ruleComponent, 1);
                //never
                default -> ruleCount = new RuleCount(ruleComponent, 1, 1);
            }
            return ruleCount;
        }
        if (jsgfRule instanceof JSGFRuleName jsgfRuleName) {
            if (jsgfRuleName.getRuleName().equals("NULL")) {
                return RuleSpecial.NULL;
            }
            if (jsgfRuleName.getRuleName().equals("VOID")) {
                return RuleSpecial.VOID;
            }
            return new RuleReference(jsgfRuleName.getRuleName());
        }
        if (jsgfRule instanceof JSGFRuleToken jsgfRuleToken) {
            return new RuleToken(jsgfRuleToken.getText());
        }
        if (jsgfRule instanceof JSGFRuleTag jsgfRuleTag) {
             RuleSequence ruleSequence=new RuleSequence();
             ruleSequence.addElement(jsgfToXml(jsgfRuleTag.getRule()));
             ruleSequence.addElement(new RuleTag(jsgfRuleTag.getTag()));
             return ruleSequence;
        }
        throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "unknown node");


    }

    public  String xmlToJSGF(String xml, boolean grammar) {
        StringBuilder xmlString = new StringBuilder();
        SrgsRuleGrammarParser parser = new SrgsRuleGrammarParser();
        writeHeaderXml(xmlString, null, grammar);
        xmlString.append(xml);
        if (!grammar) {
            xmlString.append("</grammar>");
        }

        List<Rule> rules;
        try {
            rules = parser.load(new StringReader(xmlString.toString()));

        } catch (URISyntaxException | GrammarException e) {
            throw new ResponseStatusException(HttpStatus.INTERNAL_SERVER_ERROR, "error in converting xml to jsgf");
        }

        StringBuilder jsgfString = new StringBuilder();
        if (grammar) {
            Map<String, Object> parserAtributes = parser.getAttributes();
            writeHeaderJSGF(jsgfString, (String) parserAtributes.get("name"), (String) parserAtributes.get("xml:lang"));
            Object imports = parser.getAttributes().get("imports");
            if (imports != null) {
                jsgfString.append(imports);
            }
        }

        rules.forEach(p -> jsgfString.append(p.toStringJSGF()).append("\n"));
        return jsgfString.toString();
    }

    public String jsgfToXml(String file) throws JSGFGrammarParseException, IOException, URISyntaxException, JSGFGrammarException {
        JSGFRuleGrammar grammarRaw = JSGFParser.newGrammarFromJSGF(new StringReader(file), new JSGFRuleGrammarFactory(new JSGFRuleGrammarManager()));
        StringBuilder sb = new StringBuilder();
        //todo
        writeHeaderXml(sb, null, false);
        sb.append("<meta name=\"Source\" content=\"").append(grammarRaw.getName()).append(".jsgf\"/>\n");
        for (JSGFRuleName importName : grammarRaw.getImports()) {
            sb.append("<meta name=\"seeAlso\" content=\"").append(importName.getRuleName()).append("\"/>\n");
        }


        for (String ruleName : grammarRaw.getRuleNames()) {
            JSGFRule rule = grammarRaw.getRule(ruleName);
            boolean publicRule = grammarRaw.isRulePublic(ruleName);
            writeXmlRule(sb, publicRule, ruleName, jsgfToXml(rule));
        }
        sb.append("</grammar>");

        return sb.toString();
    }

    private void writeXmlRule(StringBuilder sb, boolean publicRule, String name, RuleComponent ruleComponent) {
        sb.append("<rule ");
        if (publicRule) {
            sb.append("scope=\"public\" ");
        }
        sb.append("id=\"");
        sb.append(name).append("\">\n");
        sb.append(ruleComponent.toStringXML());
        sb.append("\n</rule>\n");
    }

    private  void writeHeaderXml(StringBuilder sb, String language, boolean justPrefix) {

        if (language == null) {
            language = "cz";
        }
        //xml
        sb.append("<?xml version=\"1.0\"?>\n");
        if (!justPrefix) {
            sb.append("<grammar version=\"1.0\"\n");
            sb.append("xml:lang='").append(language).append("'\n");
            sb.append("xmlns=\"http://www.w3.org/2001/06/grammar\">\n");
        }
    }

    private  void writeHeaderJSGF(StringBuilder sb, String name, String language) {
        sb.append("#JSGF V1.0").append(language == null ? "" : " " + language).append(";\n");

        sb.append("grammar ").append(name == null ? "defaulf" : name).append(";\n\n");

    }

}

