package com.jakubu9333.grammarManager.unitTest;

import com.jakubu9333.grammarManager.FileType;
import com.jakubu9333.grammarManager.dao.XmlDataDao;
import com.jakubu9333.grammarManager.service.FileService;
import com.jakubu9333.grammarManager.service.GrammarService;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.springframework.http.HttpStatus;
import org.springframework.web.server.ResponseStatusException;

import static org.assertj.core.api.Assertions.assertThat;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyBoolean;
import static org.mockito.ArgumentMatchers.eq;
import static org.mockito.Mockito.when;

/**
 * @author Jakub Uhlarik
 */
public class GrammarServiceTest {

    @Mock
    private FileService fileService;

    @Mock
    private XmlDataDao xmlDataDao;
    private GrammarService grammarService;

    @BeforeEach
    public void setUp() {
        MockitoAnnotations.openMocks(this);
        grammarService = new GrammarService(xmlDataDao, fileService);
    }

    @Test
    public void testEmptyRoot() {

        when(xmlDataDao.getByRoot("empty")).thenReturn("");
        try {
            grammarService.getGrammarByRootSymbol("empty");
        } catch (ResponseStatusException e) {
            assertThat(e.getStatusCode()).isEqualTo(HttpStatus.NOT_FOUND);
            assertThat(e.getReason()).isEqualTo("grammar with root rule: empty not found");
        }
    }

    @Test
    public void testRoot() {
        String xml = "<grammar root='xml'></grammar>";
        when(xmlDataDao.getByRoot("xml")).thenReturn(xml);
        String result = grammarService.getGrammarByRootSymbol("xml");
        assertThat(result).isEqualTo(xml);
    }

    @Test
    public void testEmptyName() {

        when(xmlDataDao.getByJsgfName("name")).thenReturn("");
        try {
            grammarService.getGrammarByJsgfName("name");
        } catch (ResponseStatusException e) {
            assertThat(e.getStatusCode()).isEqualTo(HttpStatus.NOT_FOUND);
            assertThat(e.getReason()).isEqualTo("grammar with name: name not found");
        }
    }

    @Test
    public void testEmptyUri() {

        when(xmlDataDao.getByUri("uri")).thenReturn("");
        try {
            grammarService.getGrammarByUri("uri", FileType.XML_SRGS);
        } catch (ResponseStatusException e) {
            assertThat(e.getStatusCode()).isEqualTo(HttpStatus.NOT_FOUND);
            assertThat(e.getReason()).isEqualTo("grammar with uri: uri not found");
        }
    }

    @Test
    public void testUriXml() {
        String xml = "<grammar r></grammar>";
        when(xmlDataDao.getByUri("xml")).thenReturn(xml);
        String result = grammarService.getGrammarByUri("xml",FileType.XML_SRGS);
        assertThat(result).isEqualTo(xml);
    }

    @Test
    public void testUriJsgf() {
        String xml = "<grammar ></grammar>";
        String jsgf= "grammar xd;";
        when(xmlDataDao.getByUri("jsgf")).thenReturn(xml);
        when(fileService.xmlToJSGF(eq(xml),anyBoolean())).thenReturn(jsgf);
        String result = grammarService.getGrammarByUri("jsgf",FileType.JSGF);
        assertThat(result).isEqualTo(jsgf);
    }

    @Test
    public void testName() {
        String xml = "<grammar ></grammar>";
        String jsgf= "grammar xd;";
        when(xmlDataDao.getByJsgfName("jsgf")).thenReturn(xml);
        when(fileService.xmlToJSGF(eq(xml),anyBoolean())).thenReturn(jsgf);
        String result = grammarService.getGrammarByJsgfName("jsgf");
        assertThat(result).isEqualTo(jsgf);
    }

    @Test
    public void testAdd() {
        String uri = "1.xml";
        String xml="<grammar></grammar>";
        when(xmlDataDao.add(xml)).thenReturn(uri);
        when(fileService.getParsedFile(eq(xml),any())).thenReturn(xml);
        String result = grammarService.addGrammarFromFile(xml,null);
        assertThat(result).isEqualTo(uri);
    }


}
